
/*################################################################################################
 *##################                                                            ##################
 *##################          XCL file loading and execution utilities          ##################
 *##################                                                            ##################
 *##################                      Implementation                        ##################
 *##################                                                            ##################
 *################################################################################################
 */


/*
 * XCLstd: Interpreter for the XCLE language, based on the libXCLstd
 * builtins implementation
 *
 * Copyright (C) 2000 Yann LANDRIN-SCHWEITZER
 *
 *
 * This library is free software;  you can redistribute it and/or modify it
 * under  the terms of the  GNU Lesser General Public License  as published
 * by the  Free Software Foundation;  either version 2.1 of the License, or
 * (at your option) any later version.
 *
 * This  library  is  distributed  in  the hope that it will be useful, but
 * WITHOUT   ANY   WARRANTY;   without   even   the   implied  warranty  of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library;  if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 *
 * Author:   Yann LANDRIN-SCHWEITZER
 * Contact:  varkhan@free.fr
 * Homepage: http://varkhan.free.fr/
 */



#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <stdlib.h>
#include <errno.h>
#include <stdio.h>
#include <fcntl.h>
#include <string.h>

#include <DbgLog.h>

#include <XCLE/Splash.h>

#include <XCLE/Containers.h>

#include <XCLE/Obj_generic.h>
#include <XCLE/Obj_void.h>
#include <XCLE/Obj_intg.h>
#include <XCLE/Obj_fltp.h>
#include <XCLE/Obj_strg.h>
#include <XCLE/Obj_list.h>
#include <XCLE/Obj_code.h>

#include <XCLE/Mod_exec.h>
#include <XCLE/Mod_fileIO.h>
#include <XCLE/Mod_parse.h>
#include <XCLE/Mod_write.h>


#include <xclutils.h>

//////////////////////////////////////////////////////////////////////////////////////////////////
/////
/////   File loading and execution
/////


// This is a HACK, the blank func was implemented in the XCLE Parse Module...
short _XCLE_isblank(char chr) ;


// Exception writer
unsigned long expprint(unsigned long lev, char * src, unsigned long idt, char * msg, void * dat) {
	        return fprintf(stderr,"[%02lu] <%s> #%04lu \"%s\"\n",lev,src,idt,msg) ;
}

// Object writer
unsigned long outprint(void * dat, char * chr, unsigned long len) {
	return sizeof(char)*fwrite(chr,sizeof(char),len,(FILE*)dat) ;
}

XCLE_WriteFormat outformat = { 0, 10, 10, 1, 52 } ;


unsigned long FileExec(char * file, unsigned long verb, XCLE_Stack stk, XCLE_Hash hsh, XCLE_ParseCtx pctx, XCLE_ExecCtx ectx) {
	unsigned long off = 0 ;
	int fd ;
	int eof, eol ;
	unsigned long ln ;       // line number
	unsigned long buffallc ; // allocated size of buffer
	unsigned long buffused ; // really used size
	unsigned long buffline ; // line length
	unsigned long buffrd ;   // read amount
	char * buffer ;
	char * tbuffer ;

	XCLE_Exception exp ;
	XCLE_List exe = NULL ;

	
	if( (file[0]=='-') && (file[1]==0) ) fd = dup(0) ;
	else fd = open(file,O_RDONLY) ;
	if(fd==-1) {
		fprintf(stderr,"Open failed on %s; aborting\n",file) ;
		return -1 ;
	}
	eof = 0 ;
	ln = 0 ;
	buffallc = 1 ;
	buffused = 0 ;
	buffline = 0 ;
	buffer = malloc(buffallc+1) ;
	if(buffer==NULL) {
		fprintf(stderr,"Failed allocation; aborting\n")	;
		return -1 ;
	}
	while( (!eof) || (buffused>0) ) {
		eol = 0 ;
		// line read into buffer
		while(!eol) {
			buffrd = read(fd,buffer+buffused,buffallc-buffused) ;
			if(buffrd==-1) {
				fprintf(stderr,"Read error in %s, line %lu; aborting\n",file,ln) ;
				return -1 ;
			}
			buffused += buffrd ;
			if(buffused<buffallc) {
				eof = 1 ;
				eol = 1 ;
				if( (buffused==0) || (buffer[buffused-1]!='\n') ) {
					buffer[buffused]='\n' ;
					buffused ++ ;
				}
			}
			else {
				tbuffer = realloc(buffer,(buffallc<<1)+1) ;
				if(tbuffer==NULL) {
					fprintf(stderr,"Failed allocation; aborting\n") ;
					return -1 ;
				}
				buffer = tbuffer ;
				buffallc = (buffallc<<1) ;
			}
		}
		buffline = 0 ;
		while(buffline<buffused) {
			if(buffer[buffline]=='\n') {
				ln ++ ;
				// empty line ?
				if(buffline==0) {
					eol = 1 ;
					buffline ++ ;
					break ;
				}
				// treat escape sequence, one less char
				else if(buffer[buffline-1]=='\\') {
					off = buffline ;
					while(off<buffused) {
						buffer[off-1] = buffer[off] ;
						off ++ ;
					}
					buffused -- ;
				}
				// real end of line
				else {
					eol = 1 ;
					buffline ++ ;
					break ;
				}
			}
			buffline ++ ;
		}
		// line parsing
		buffer[buffline-1] = 0 ;
		if(buffer[0]=='#') {
			if( (buffer[1]=='#') && (verb&PRINT_COM) ) {
				fprintf(stderr,">>%s\n",buffer+2) ;
			}
			off = 0 ;
			while(off<buffused-buffline) {
				buffer[off] = buffer[off+buffline] ;
				off ++ ;
			}
			buffused -= buffline ;
			continue ;
		}
		off = 0 ;
		while(_XCLE_isblank(buffer[off])) {
			if(buffer[off]==0) break ;
			off ++ ;
		}
		if(buffer[off]==0) {
			off = 0 ;
			while(off<buffused-buffline) {
				buffer[off] = buffer[off+buffline] ;
				off ++ ;
			}
			buffused -= buffline ;
			continue ;
		}
		errno = 0 ;
		exe = XCLE_ListParse(pctx,buffer,&off) ;
		// line evaluation
		if( (exe==NULL) && (errno==EINVAL) ) {
			fprintf(stderr,"Syntax error in %s, line %lu: %50.50s...\n",file,ln,buffer) ;
			if(off<20) {
				fprintf(stderr,"%-80.80s\n",buffer) ;
				fprintf(stderr,"%*.*s^\n",(int)off,(int)off,"                    ") ;
			}
			else {
				fprintf(stderr,"...%20.20s%-60.60s\n",buffer+off-20,buffer+off) ;
				fprintf(stderr,"   %20.20s^\n","                    ") ;
			}
			off = 0 ;
			while(off<buffused-buffline) {
				buffer[off] = buffer[off+buffline] ;
				off ++ ;
			}
			buffer[off] = 0 ;
			buffused -= buffline ;
			continue ;
		}
		XCLE_ListUpRef(exe) ;
		errno = 0 ;
		exp = XCLE_ListEval(ectx,stk,hsh,exe) ;
		if(exp==NULL) {
			fprintf(stderr,"Execution failed ; aborting\n") ;
			return -1 ;
		}
		if(verb&REPORT_ALL) {
			fprintf(stderr,"\nEvaluated ") ;
			fflush(stderr) ;
			XCLE_ListWrite(exe,outprint,stderr,outformat) ;
			if(exp!=XCLE_EXCEPTION_OK) {
				fprintf(stderr," ; Exception at:\n") ;
				XCLE_ExecCtxMapExceptions(ectx,expprint,NULL) ;
			}
			else fprintf(stderr," ; OK\n") ;
			fprintf(stderr,"\n") ;
		}
		if(verb&STACK_ALL) {
			XCLE_StackWrite(stk,outprint,stderr,outformat,STK_HDR,STK_FMT,STK_TLR,8) ;
		}
		if(verb&PRINT_ALL) {
			if(XCLE_StackDepth(stk)>0) XCLE_ObjectWrite(XCLE_StackGet(stk,1),outprint,stdout,outformat) ;
			printf("\n") ;
		}
		XCLE_ExecCtxClear(ectx) ;
		XCLE_ListDnRef(exe) ;
		XCLE_ListFree(exe) ;
		off = 0 ;
		while(off<buffused-buffline) {
			buffer[off] = buffer[off+buffline] ;
			off ++ ;
		}
		buffused -= buffline ;
	}
	free(buffer) ;
	close(fd) ;
	return ln ;
}



unsigned long StrExec(char * str, unsigned long verb, XCLE_Stack stk, XCLE_Hash hsh, XCLE_ParseCtx pctx, XCLE_ExecCtx ectx) {
	unsigned long off = 0 ;
	XCLE_Exception exp ;
	XCLE_List exe = NULL ;
	errno = 0 ;
	while(_XCLE_isblank(str[off])) {
		if(str[off]==0) break ;
		off ++ ;
	}
	if(str[off]==0) {
		fprintf(stderr,"\nEmpty string\n") ;
		return -1 ;
	}
	exe = XCLE_ListParse(pctx,str,&off) ;
	if( (exe==NULL) && (errno==EINVAL) ) {
		fprintf(stderr,"Syntax error : %50.50s...\n",str) ;
		if(off<20) {
			fprintf(stderr,"%-80.80s\n",str) ;
			fprintf(stderr,"%*.*s^\n",(int)off,(int)off,"                    ") ;
		}
		else {
			fprintf(stderr,"...%20.20s%-60.60s\n",str+off-20,str+off) ;
			fprintf(stderr,"   %20.20s^\n","                    ") ;
		}
		return -1 ;
	}
	XCLE_ListUpRef(exe) ;
	exp = XCLE_ListEval(ectx,stk,hsh,exe) ;
	if(exp==NULL) {
		fprintf(stderr,"Execution failed ; aborting\n") ;
		return -1 ;
	}
	if(verb&REPORT_ALL) {
		fprintf(stderr,"\nEvaluated ") ;
		fflush(stderr) ;
		XCLE_ListWrite(exe,outprint,stderr,outformat) ;
		if(exp!=XCLE_EXCEPTION_OK) {
			fprintf(stderr," ; Exception at:\n") ;
			XCLE_ExecCtxMapExceptions(ectx,expprint,NULL) ;
		}
		else fprintf(stderr," ; OK\n") ;
		fprintf(stderr,"\n") ;
	}
	if(verb&STACK_ALL) {
		XCLE_StackWrite(stk,outprint,stderr,outformat,STK_HDR,STK_FMT,STK_TLR,8) ;
	}
	if(verb&PRINT_ALL) {
		if(XCLE_StackDepth(stk)>0) XCLE_ObjectWrite(XCLE_StackGet(stk,1),outprint,stdout,outformat) ;
		printf("\n") ;
	}
	XCLE_ExecCtxClear(ectx) ;
	XCLE_ListDnRef(exe) ;
	XCLE_ListFree(exe) ;
	return 1 ;
}

